/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 *
 * Portions created by or assigned to Jabber.com, Inc. are
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * iq.c - Handles iq for a session.  register, search, time, etc.
 * -------------------------------------------------------------------------- */

#include "msntrans.h"
#include <sys/utsname.h>

void mt_iq_reg_get(session s, jpacket jp);
void mt_iq_reg_set(session s, jpacket jp);
void mt_iq_gateway_get(session s, jpacket jp);
void mt_iq_gateway_set(session s, jpacket jp);
void mt_iq_browse_user(session s, jpacket jp);

void mt_iq(session s, jpacket jp)
{
    char *ns;

    ns = xmlnode_get_attrib(jp->iq,"xmlns");

    switch (jpacket_subtype(jp))
    {
    case JPACKET__GET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
            mt_iq_reg_get(s,jp);
        else if (j_strcmp(ns,NS_SEARCH) == 0)
            mt_iq_search_get(s,jp);
        else if (j_strcmp(ns,NS_VERSION) == 0)
            mt_iq_version(s->ti,jp);
        else if (j_strcmp(ns,NS_TIME) == 0)
            mt_iq_time(s->ti,jp);
        else if (j_strcmp(ns,NS_VCARD) == 0)
        {
            if (jp->to->user == NULL)
                mt_iq_vcard_server(s->ti,jp);
            else
                xmlnode_free(jp->x);
        }
        else if (j_strcmp(ns,NS_LAST) == 0)
        {
            if (jp->to->user == NULL)
                xmlnode_insert_node(jutil_iqresult(jp->x),s->ti->last);
            else
                jutil_error(jp->x,TERROR_NOTALLOWED);

            mt_deliver(s->ti,jp->x);
        }
        else if (j_strcmp(ns,NS_BROWSE) == 0)
            jp->to->user == NULL ? mt_iq_browse_server(s->ti,jp) : mt_iq_browse_user(s,jp);
        else if (j_strcmp(ns,NS_GATEWAY) == 0)
            mt_iq_gateway_get(s,jp);
        else
        {
            jutil_error(jp->x,TERROR_NOTIMPL);
            mt_deliver(s->ti,jp->x);
        }
        break;
    case JPACKET__SET:
        if (j_strcmp(ns,NS_REGISTER) == 0)
            mt_iq_reg_set(s,jp);
        else if (j_strcmp(ns,NS_SEARCH) == 0)
            mt_iq_search_set(s,jp);
        else
        {
            if (j_strcmp(ns,NS_VERSION) == 0 || j_strcmp(ns,NS_TIME) == 0)
                jutil_error(jp->x,TERROR_NOTALLOWED);
            else
                jutil_error(jp->x,TERROR_NOTIMPL);  
            mt_deliver(s->ti,jp->x);
        }
        break;
    default:
        xmlnode_free(jp->x);
        break;
    }
}

void mt_iq_reg_get(session s, jpacket jp)
{
    mti ti = s->ti;
    xmlnode q, reg;

    if (s->type == stype_register)
    {
        jutil_iqresult(jp->x);
        mt_deliver(ti,jp->x);
        return;
    }

    reg = xdb_get(ti->xc,mt_xdb_id(xmlnode_pool(jp->x),s->id,s->server),NS_REGISTER);
    if (reg != NULL)
    {
        jutil_iqresult(jp->x);
        q = xmlnode_insert_tag(jp->x,"query");
        xmlnode_put_attrib(q,"xmlns",NS_REGISTER);

        /* don't send back the password */
        xmlnode_hide(xmlnode_get_tag(reg,"password"));
        xmlnode_insert_node(q,xmlnode_get_firstchild(reg));
        xmlnode_free(reg);

        xmlnode_insert_tag(q,"password");
        xmlnode_insert_tag(q,"registered");
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"instructions"),ti->reg,-1);
    }
    else
        jutil_error(jp->x,TERROR_INTERNAL);

    mt_deliver(ti,jp->x);
}

void mt_iq_reg_set(session s, jpacket jp)
{
    mti ti = s->ti;
    jid id;
    char *key, *nick;

    log_debug(ZONE,"Session[%s], register request",jid_full(s->id));

    if (s->type == stype_register)
    {
        xmlnode_free(jp->x);
        return;
    }

    key = xmlnode_get_tag_data(jp->iq,"key");
    if (key == NULL || jutil_regkey(key,jid_full(jp->from)) == NULL)
    {
        log_debug(ZONE,"Session[%s], register failed, invalid key",jid_full(s->id));
        jutil_error(jp->x,TERROR_NOTACCEPTABLE);
        mt_deliver(ti,jp->x);
        return;
    }

    id = mt_xdb_id(xmlnode_pool(jp->x),s->id,s->server);
    if (xmlnode_get_tag(jp->iq,"remove") != NULL)
    {
        log_debug(ZONE,"Session[%s], unregistering",jid_full(s->id));

        xdb_set(ti->xc,id,NS_REGISTER,NULL);
        xdb_set(ti->xc,id,NS_ROSTER,NULL);

        jutil_iqresult(jp->x);
        mt_deliver(s->ti,jp->x);
        mt_deliver(s->ti,mt_presnew(JPACKET__UNSUBSCRIBE,jid_full(s->id),spools(s->p,s->server,"/registered",s->p),NULL));
        mt_session_end(s);
        return;
    }

    nick = mt_safe_nick(s->p,xmlnode_get_tag_data(jp->iq,"nick"));
    if (nick != NULL)
    {
        xmlnode_hide(xmlnode_get_tag(jp->iq,"username"));
        xmlnode_hide(xmlnode_get_tag(jp->iq,"password"));
        xmlnode_hide(xmlnode_get_tag(jp->iq,"instructions"));

        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"username"),s->user,-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"password"),s->pass,-1);

        if (xdb_set(ti->xc,id,NS_REGISTER,xmlnode_dup(jp->iq)) == 0)
        {
            if (j_strcmp(s->nick,nick) != 0)
            {
                s->nick = nick;
                mt_mpstream_write(s->st,mt_cmd_rea(s->st->trid,s->user,nick));
            }
            jutil_iqresult(jp->x);
        }
        else
            jutil_error(jp->x,TERROR_UNAVAIL);
        mt_deliver(ti,jp->x);
        return;
    }

    log_debug(ZONE,"One or more registration fields missing or invalid");

    jutil_error(jp->x,TERROR_NOTACCEPTABLE);
    mt_deliver(ti,jp->x);
}

void mt_iq_search_get(session s, jpacket jp)
{
    xmlnode q;

    jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(jp->x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_SEARCH);
    xmlnode_insert_tag(q,"first");
    xmlnode_insert_tag(q,"last");

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"key"),jutil_regkey(NULL,jid_full(jp->from)),-1);

    mt_deliver(s->ti,jp->x);
}

#define cdata2fnd(p,name,data) spools(p,name,"=",data ? data : "*",p)

void mt_iq_search_set(session s, jpacket jp)
{
    pool p = jp->p;
    xmlnode q = jp->iq;
    char *f, *l, *key, *first, *last;

    key = xmlnode_get_tag_data(q,"key");
    first = xmlnode_get_tag_data(q,"first");
    last = xmlnode_get_tag_data(q,"last");

    if (key == NULL || jutil_regkey(key,jid_full(jp->from)) == NULL || (first== NULL && last == NULL))
    {
        jutil_error(jp->x,TERROR_NOTACCEPTABLE);
        mt_deliver(s->ti,jp->x);
        return;
    }

    if (s->connected == 0)
    {
        jpq jq;
        log_debug(ZONE,"Queued");
        jq = pmalloco(p,sizeof(_jpq));
        jq->jp = jp;
        pth_msgport_put(s->queue,(pth_message_t *)jq);
        return;
    }

    f = cdata2fnd(p,"fname",last);
    l = cdata2fnd(p,"lname",last);

    /* jabber:iq:search doesn't have these */
/*  c = cdata2fnd(p,"city",NULL);
    st = cdata2fnd(p,"state",NULL);
    co = cdata2fnd(p,"country",NULL);
*/
    jutil_iqresult(jp->x);
    xmlnode_put_vattrib(jp->x,"s",(void*) s);
    mt_mpstream_rwrite(s->st,(handle) mt_phandle_fnd,mt_cmd_fnd(s->st->trid,f,l,NULL,NULL,NULL),(void*) jp->x);
}

void mt_iq_version(mti ti, jpacket jp)
{
    xmlnode os, x, q;
    struct utsname un;

    x = jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_VERSION);

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"name"),"MSN Transport",-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"version"),"1.1",-1);

    uname(&un);
    os = xmlnode_insert_tag(q,"os");
    xmlnode_insert_cdata(os,un.sysname,-1);
    xmlnode_insert_cdata(os," ",1);
    xmlnode_insert_cdata(os,un.release,-1);

    mt_deliver(ti,x);
}

void mt_iq_time(mti ti, jpacket jp)
{
    xmlnode x, q;

    x = jutil_iqresult(jp->x);

    q = xmlnode_insert_tag(x,"query");
    xmlnode_put_attrib(q,"xmlns",NS_TIME);

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"utc"),jutil_timestamp(),-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"tz"),tzname[0],-1);

    mt_deliver(ti,x);
}

void mt_iq_vcard_server(mti ti, jpacket jp)
{
    /* send back transport vCard */
    xmlnode_insert_node(jutil_iqresult(jp->x),ti->vcard);
    mt_deliver(ti,jp->x);
}

void mt_iq_gateway_get(session s, jpacket jp)
{
    if (jp->to->user == NULL)
    {
        xmlnode q;

        jutil_iqresult(jp->x);

        q = xmlnode_insert_tag(jp->x,"query");
        xmlnode_put_attrib(q,"xmlns",NS_GATEWAY);
        xmlnode_insert_cdata(xmlnode_insert_tag(q,"desc"),"Enter the user's MSN account",-1);
        xmlnode_insert_tag(q,"prompt");
    }
    else
        jutil_error(jp->x,TERROR_NOTALLOWED);

    mt_deliver(s->ti,jp->x);
}

void mt_iq_gateway_set(session s, jpacket jp)
{
    char *user;

    user = xmlnode_get_tag_data(jp->iq,"prompt");

    if (user)
    {
        char *ptr;

        ptr = strchr(user,'@');
        if (ptr)
        {
            xmlnode q;

            *ptr = '%';

            jutil_iqresult(jp->x);
            q = xmlnode_insert_tag(jp->x,"query");
            xmlnode_put_attrib(q,"xmlns",NS_GATEWAY);
            xmlnode_insert_cdata(xmlnode_insert_tag(q,"prompt"),user,-1);
            mt_deliver(s->ti,jp->x);

            return;
        }
    }

    jutil_error(jp->x,TERROR_BAD);
    mt_deliver(s->ti,jp->x);
}

void mt_iq_browse_server(mti ti, jpacket jp)
{
    xmlnode q;

    q = xmlnode_insert_tag(jutil_iqresult(jp->x),"service");
    xmlnode_put_attrib(q,"xmlns",NS_BROWSE);

    xmlnode_put_attrib(q,"type","msn");
    xmlnode_put_attrib(q,"jid",jp->to->server);
    xmlnode_put_attrib(q,"name",xmlnode_get_tag_data(ti->vcard,"FN"));

    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_REGISTER,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_SEARCH,-1);
    xmlnode_insert_cdata(xmlnode_insert_tag(q,"ns"),NS_GATEWAY,-1);

    mt_deliver(ti,jp->x);
}

void mt_iq_browse_user(session s, jpacket jp)
{
    xmlnode browse;
    muser u;
    char *m;

    if (s->type == stype_register)
    {
        jpq jq;

        jq = pmalloco(jp->p,sizeof(_jpq));
        jq->jp = jp;
        pth_msgport_put(s->queue,(pth_message_t *)jq);
        return;
    }

    m = mt_jid2mid(jp->p,jp->to);
    if (m == NULL)
    {
        jutil_error(jp->x,TERROR_BAD);
        mt_deliver(s->ti,jp->x);
        return;
    }

    u = (muser) ghash_get(s->muser_tab,m);

    jutil_iqresult(jp->x);
    browse = xmlnode_insert_tag(jp->x,"user");
    xmlnode_put_attrib(browse,"jid",jid_full(jid_user(jp->to)));
    xmlnode_put_attrib(browse,"type","user");

    if (u && u->handle)
        xmlnode_put_attrib(browse,"name",u->handle);

    mt_deliver(s->ti,jp->x);
}
