/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 *
 * Portions created by or assigned to Jabber.com, Inc. are
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * -------------------------------------------------------------------------- */

#include "msntrans.h"

void mt_session_cleanup(void *arg);
void _mt_session_end(void *arg);

session mt_session_create(mti ti, jpacket jp)
{
    pool p;
    session s;
    jid id = jp->from;

    log_debug(ZONE,"Creating session for %s",jid_full(id));

    p = pool_new();
    s = pmalloco(p,sizeof(_session));
    s->p = p;
    s->ti = ti;
    s->q = mtq_new(p);

    s->id = jid_new(p,jid_full(jid_user(id)));
    s->state = ustate_fln;
    s->from = jid_new(p,jid_full(jp->to));
    jid_set(s->from,"registered",JID_RESOURCE);
    s->server = s->from->server;
    s->queue = pth_msgport_create("mt_session_queue");
    s->type = stype_normal;

    ghash_put(ti->sessions,jid_full(s->id),s);
    ++ti->sessions_count;

    return s;
}

void mt_session_connect(session s, char *host, int port)
{
    mpstream st;

    s->st = NULL;
    s->st = st = mt_mpstream_connect(s,host,port,mt_phandle_default,(void *) s);
    pool_cleanup(st->p,mt_session_cleanup,(void *) st);
    mt_mpstream_rwrite(st,mt_phandle_ver,mt_cmd_ver(st->trid),(void *) s);
}

void mt_session_cleanup(void *arg)
{
    mpstream st = (mpstream) arg;
    session s = st->s;

    log_debug(ZONE,"Session[%s], connection to MSN server closed, %X %X",jid_full(s->id),st,s->st);

    if (st == s->st)
    {
        s->st = NULL;
        if (s->exit_flag == 0);
            mt_session_start(s);
    }
}

void mt_session_start(session s)
{
    mti ti = s->ti;

    if (ti->shutdown)
    {
        mt_session_end(s);
        return;
    }

    s->connected = 0;

    /* Have we gone over the maximum DS connection attempt limit? */
    if (s->conn_at < CONN_AT_MAX)
    {
        ++s->conn_at;
        mt_session_connect(s,ti->server,1863);
        return;
    }

    log_debug(ZONE,"Session[%s], DS connection attempts reached max",jid_full(s->id));

    mt_session_kill(s,TERROR_EXTERNAL);
}

void mt_session_in(void *arg)
{
    jpacket jp = (jpacket) arg;
    session s = (session) jp->aux1;

    mt_session_jpacket(s,jp);
}

void mt_session_jpacket(session s, jpacket jp)
{
    if (s->exit_flag == 1)
    {
        xmlnode_free(jp->x);
        return;
    }

    switch (jp->type)
    {
    case JPACKET_PRESENCE:
        mt_presence(s,jp);
        break;
    case JPACKET_S10N:
        mt_s10n(s,jp);
        break;
    case JPACKET_MESSAGE:
        mt_message(s,jp);
        break;
    case JPACKET_IQ:
        mt_iq(s,jp);
        break;
    default:
        xmlnode_free(jp->x);
        break;
    }
}

void mt_session_packet(void *arg)
{
    smsg sm = (smsg) arg;
    mpstream st = (mpstream) sm->data1;
    mpacket ph = (mpacket) sm->data2;
    session s = st->s;

    if (s->exit_flag == 0)
        mt_mpstream_dispatch(st,ph);

    pool_free(ph->p);
}

void mt_session_register(session s)
{
    mti ti = s->ti;
    jpq jq;
    jpacket jp;
    pool p;
    xmlnode xreg;

    s->type = stype_normal;
    s->muser_tab = ghash_create(25,(KEYHASHFUNC)str_hash_code,(KEYCOMPAREFUNC)j_strcmp);
    s->state = ustate_nln;

    jq = (jpq) pth_msgport_get(s->queue);
    jp = jq->jp;
    p = jp->p;

    xmlnode_hide(xmlnode_get_tag(jp->iq,"instructions"));
    xreg = xmlnode_dup(jp->iq);

    if (xdb_set(ti->xc,mt_xdb_id(p,s->id,s->server),NS_REGISTER,xreg) == 0)
    {
        jutil_iqresult(jp->x);

        mt_deliver(ti,mt_presnew(JPACKET__SUBSCRIBE,jid_full(s->id),jid_full(s->from),NULL));
        mt_deliver(ti,mt_presnew(JPACKET__PROBE,jid_full(s->id),jid_full(s->from),NULL));
    }
    else
    {
        xmlnode_free(xreg);
        jutil_error(jp->x,TERROR_UNAVAIL);
        mt_session_end(s);
    }

    mt_deliver(ti,jp->x);
}

void _mt_session_end(void *arg)
{
    session s = (session) arg;
    mti ti = s->ti;
    jpq jq;

    log_debug(ZONE,"Session[%s], exiting",jid_full(s->id));

    if (s->st != NULL && s->st->state != conn_CLOSED)
    {
        if (s->connected)
            mt_mpstream_write(s->st,mt_cmd_out());

        mio_close(s->st->c);
    }

    if (s->muser_tab)
    {
        mt_deliver(ti,mt_presnew(JPACKET__UNAVAILABLE,jid_full(s->id),jid_full(s->from),NULL));
        ghash_walk(s->muser_tab,mt_muser_cleanup_walker,(void *) s);
        ghash_destroy(s->muser_tab);
    }

    while ((jq = (jpq) pth_msgport_get(s->queue)) != NULL)
    {
        jutil_error(jq->jp->x,TERROR_NOTFOUND);
        mt_deliver(s->ti,jq->jp->x);
    }

    pth_msgport_destroy(s->queue);
    ppdb_free(s->p_db);

    ghash_remove(ti->sessions,jid_full(s->id));

    s->next = ti->dead;
    ti->dead = s;

    log_debug(ZONE,"Session exited %d",s->ref);
}

void mt_session_end(session s)
{
    if (s->exit_flag == 1)
        return;

    log_debug(ZONE,"Killing session[%s]",jid_full(s->id));

    s->exit_flag = 1;

    if (s->type == stype_register)
    {
        jpq jq;
        xmlnode x;

        jq = (jpq) pth_msgport_get(s->queue);
        x = jq->jp->x;
        jutil_error(x,TERROR_NOTACCEPTABLE);
    }

    mtq_send(s->q,s->p,_mt_session_end,(void *) s);
}

void mt_session_kill(session s, terror t)
{
    if (s->exit_flag == 1)
        return;

    log_debug(ZONE,"Killing session[%s], error %s",jid_full(s->id),t.msg);

    s->exit_flag = 1;

    if (s->type == stype_register)
    {
        jpq jq;
        xmlnode x;

        jq = (jpq) pth_msgport_get(s->queue);
        x = jq->jp->x;
        jutil_error(x,t);

        mt_deliver(s->ti,x);
    }

    mtq_send(s->q,s->p,_mt_session_end,(void *) s);
}
