/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 *
 * Portions created by or assigned to Jabber.com, Inc. are
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1999-2000 Schuyler Heath <sheath@jabber.org>
 *
 * Acknowledgements
 *
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 *
 * -------------------------------------------------------------------------- */

#include "msntrans.h"

void mt_mpstream_read(mio c, int state, void *arg, char *buffer, int bufsz);
int mt_mpstream_packet(mpacket ph, void *arg);
void mt_mpstream_close(void *arg);

mpstream mt_mpstream_connect(session s, char *host, int port, handle h, void *arg)
{
    pool p;
    mpstream st;

    p = pool_new();

    st = pmalloco(p,sizeof(_mpstream));
    st->p = p;
    st->s = s;
    st->trid = 1 + (unsigned long) (15.0 * rand() / (RAND_MAX + 1.0));
    st->h = h;
    st->arg = arg;
    st->par = mt_parser_new(p,mt_mpstream_packet,(void *) st);
    st->state = conn_OPEN;

    ++s->ref;

    mio_connect(host,port,mt_mpstream_read,(void *) st,10,NULL,NULL);

    return st;
}

int mt_mpstream_write(mpstream st, mpacket ph)
{
    if (st->state == conn_CLOSED)
    {
        log_debug(ZONE,"MIO already closed");
        pool_free(ph->p);
        return 0;
    }

    if (st->c != NULL)
    {
        mio_write(st->c,NULL,mt_packet_str(ph),-1);
        pool_free(ph->p);
    }
    else
        st->buf = ph;

    ++st->trid;

    return 1;
}

/* register a new packet handler */
void mt_mpstream_register(mpstream st, handle h, unsigned long trid, void *arg)
{
    mphandler ph, cur;
    pool p;

    p = pool_new();
    ph = pmalloco(p,sizeof(_mphandler));
    ph->p = p;
    ph->trid = trid;
    ph->h = h;
    ph->arg = arg;

    if (st->phan != NULL)
    {
        for (cur = st->phan; cur->next != NULL; cur = cur->next);
        cur->next = ph;
    }
    else
        st->phan = ph;
}

void mt_mpstream_dispatch(mpstream st, mpacket ph)
{
    mphandler cur, prev;
    jid id = st->s->id;
    char *str;
    unsigned long trid = 0;

    str = mt_packet_data(ph,1);
    if (str && (trid = atol(str)))
    {
        /* look for a packet registered handler */
        cur = st->phan;
        prev = NULL;
        while (cur != NULL)
        {
            if (cur->trid == trid)
            {
                log_debug(ZONE,"Session[%s], packet handler found",jid_full(id));

                if ((cur->h)(ph,cur->arg))
                {
                    if (prev == NULL)
                        st->phan = cur->next;
                    else
                        prev->next = cur->next;
                    pool_free(cur->p);
                }
                return;
            }

            prev = cur;
            cur = cur->next;
        }
    }

    /* no handler found, call the default handler */
    if ((st->h) (ph,st->arg))
        log_debug(ZONE,"Session[%s], unknown command %s",jid_full(id),mt_packet_data(ph,0));
}

void mt_mpstream_read(mio c, int state, void *arg, char *buffer, int bufsz)
{
    mpstream st = (mpstream) arg;
    session s = st->s;

    switch (state)
    {
    case MIO_NEW:
        if (s->exit_flag || st->state == conn_CLOSED)
        {
            mio_close(c);
            return;
        }

        st->c = c;
        /* we don't want any karma */
        mio_karma(c,KARMA_INIT,KARMA_MAX,KARMA_INC,0,KARMA_PENALTY,KARMA_RESTORE);

        if (st->buf)
        {
            mpacket ph = st->buf;
            st->buf = NULL;

            mio_write(c,NULL,mt_packet_str(ph),-1);
            pool_free(ph->p);
        }
        break;

    case MIO_BUFFER:
        log_debug(ZONE,"Session[%s], read %d bytes",jid_full(s->id),bufsz);
        if (s->exit_flag == 0) mt_packet_parse(st->par,buffer);
        break;

    case MIO_CLOSED:
    case MIO_ERROR:
        if (st->state != conn_CLOSED)
        {
            st->state = conn_CLOSED;
            mtq_send(s->q,st->p,mt_mpstream_close,(void*) st);
        }
        break;
    }
}

/* packet parsing callback */
int mt_mpstream_packet(mpacket ph, void *arg)
{
    mpstream st = (mpstream) arg;
    session s = st->s;
    smsg sm;

    if (ph == NULL)
    {
        log_error(s->server,"Session[%s], error parsing MSN packet",jid_full(s->id));
        mio_close(st->c);
        return 0;
    }

    if (s->exit_flag)
    {
        pool_free(ph->p);
        return 0;
    }

#ifdef PACKET_DEBUG
    log_debug(ZONE,"Session[%s], received packet\n%s",jid_full(s->id),mt_packet_str(ph));
#endif

    sm = pmalloco(ph->p,sizeof(_smsg));
    sm->data1 = st;
    sm->data2 = ph;

    /* send it to the session */
    mtq_send(s->q,ph->p,mt_session_packet,(void*) sm);

    return 1;
}

void mt_mpstream_close(void *arg)
{
    mpstream st = (mpstream) arg;
    session s = st->s;
    mphandler temp, cur = st->phan;

    log_debug(ZONE,"closed");

    /* free the packet handlers */
    while (cur != NULL)
    {
        temp = cur;
        cur = cur->next;
        pool_free(temp->p);
    }

    if (st->buf)
        pool_free(st->buf->p);

    pool_free(st->p);
    --s->ref;
}
